#if !defined(TORCH_STABLE_ONLY) && !defined(TORCH_TARGET_VERSION)
#pragma once

// @generated by torchgen/gen.py from Function.h

#include <ATen/Context.h>
#include <ATen/DeviceGuard.h>
#include <ATen/TensorUtils.h>
#include <ATen/TracerMode.h>
#include <ATen/core/Generator.h>
#include <ATen/core/Reduction.h>
#include <ATen/core/Tensor.h>
#include <c10/core/Scalar.h>
#include <c10/core/Storage.h>
#include <c10/core/TensorOptions.h>
#include <c10/util/Deprecated.h>
#include <optional>
#include <string_view>



#include <ATen/ops/_foreach_clone_ops.h>

namespace at {


// aten::_foreach_clone(Tensor[] self, *, MemoryFormat? memory_format=None) -> Tensor[]
inline ::std::vector<at::Tensor> _foreach_clone(at::TensorList self, ::std::optional<at::MemoryFormat> memory_format=::std::nullopt) {
    return at::_ops::_foreach_clone::call(self, memory_format);
}

// aten::_foreach_clone.out(Tensor[] self, *, MemoryFormat? memory_format=None, Tensor(a!)[] out) -> ()
inline void _foreach_clone_out(at::TensorList out, at::TensorList self, ::std::optional<at::MemoryFormat> memory_format=::std::nullopt) {
    return at::_ops::_foreach_clone_out::call(self, memory_format, out);
}
// aten::_foreach_clone.out(Tensor[] self, *, MemoryFormat? memory_format=None, Tensor(a!)[] out) -> ()
inline void _foreach_clone_outf(at::TensorList self, ::std::optional<at::MemoryFormat> memory_format, at::TensorList out) {
    return at::_ops::_foreach_clone_out::call(self, memory_format, out);
}

}

#else
#error "This file should not be included when either TORCH_STABLE_ONLY or TORCH_TARGET_VERSION is defined."
#endif  // !defined(TORCH_STABLE_ONLY) && !defined(TORCH_TARGET_VERSION)
